"""This module provides the client class for interacting with a SQLite database."""
from . import Base, PaginateParameters, Partial, ResourceId, ResourceObj


class SQLiteClient(Base):
    """A class representing a SQLite DB client for database operations.

    Args:
        path (str): The path to the SQLite database.
        table (str): The name of the table.

    Raises:
        self.exception: If the resource is not found in the database.
    """

    def __init__(self, path: str, table: str) -> None:
        """Initialize the SQLiteClient class.

        Args:
            path (str): The path to the database.
            table (str): The name of the table in the database.

        """

    async def insert_one(self, resource_id: ResourceId, resource_obj: ResourceObj) -> ResourceObj:
        """Inserts a resource object in the database.

        Args:
            resource_id (ResourceId): The ID of the resource to be inserted.
            resource_obj (ResourceObj): The resource object to be inserted.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource already exists in the database.
        """
        raise NotImplementedError

    async def select_one(self, resource_id: ResourceId) -> ResourceObj:
        """Retrieves a resource from the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be retrieved.

        Returns:
            ResourceObj: The retrieved resource object.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        raise NotImplementedError

    async def update_one(
        self, resource_id: ResourceId, resource_obj: ResourceObj, partial: Partial = None
    ) -> ResourceObj:
        """Updates a resource in the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be updated.
            resource_obj (ResourceObj): The resource object to be updated.
            partial (Partial): Whether to perform a partial update or replace the entire resource object.
                Defaults to None.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        raise NotImplementedError

    async def delete_one(self, resource_id: ResourceId) -> None:
        """Deletes a resource from the database based on the given resource ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be deleted.

        Returns:
            None

        Raises:
            self.exception: If the resource is not found in the database.
        """
        raise NotImplementedError

    async def select_many(self, paginate_parameters: PaginateParameters) -> list[ResourceObj]:
        """Retrieves multiple resources from the database based on the given pagination parameters.

        Args:
            paginate_parameters (PaginateParameters): The pagination parameters.

        Returns:
            list[ResourceObj]: A list of objects representing the retrieved resources.
        """
        raise NotImplementedError



"""This module provides the dependencies for sthali-db usage."""
from typing import Annotated, NoReturn

from pydantic import BaseModel, Field, NonNegativeInt


async def filter_parameters() -> NoReturn:
    """Not implemented."""
    raise NotImplementedError


class PaginateParameters(BaseModel):
    """Represents the parameters for retrieving items.

    Attributes:
        skip (NonNegativeInt): The number of items to skip. Defaults to 0.
        limit (NonNegativeInt): The maximum number of items to return. Defaults to 100.
    """

    skip: Annotated[NonNegativeInt, Field(default=0, description="The number of items to skip")]
    limit: Annotated[NonNegativeInt, Field(default=100, description="The maximum number of items to return")]



from unittest import IsolatedAsyncioTestCase

from sthali_db.dependencies import filter_parameters, PaginateParameters


class TestFilterParameters(IsolatedAsyncioTestCase):
    async def test_not_implemented(self) -> None:
        with self.assertRaises(NotImplementedError):
            await filter_parameters()


class TestPaginateParameters(IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        result = PaginateParameters()  # type: ignore

        self.assertEqual(result.skip, 0)
        self.assertEqual(result.limit, 100)

    async def test_return_custom(self) -> None:
        result = PaginateParameters(skip=10, limit=10)

        self.assertEqual(result.skip, 10)
        self.assertEqual(result.limit, 10)



"""This module provides the necessary components for interacting with the database."""
from . import clients, dependencies, models
from .clients import DB, DBSpecification
from .dependencies import PaginateParameters, filter_parameters
from .models import Base, BaseWithId, FieldDefinition, Models

__all__ = [
    # "Base",
    # "BaseWithId",
    "DB",
    "DBSpecification",
    "FieldDefinition",
    "Models",
    "PaginateParameters",
    # "clients",
    # "dependencies",
    # "filter_parameters",
    "models",
]



"""This module provides classes for creating dynamic models based on field definitions."""
from collections.abc import Callable
from typing import Annotated, Any, TypeVar
from uuid import UUID

from pydantic import BaseModel, Field, create_model  # type: ignore
from pydantic.dataclasses import dataclass


@dataclass
class Default:
    """Represents a default value for an attribute.

    Attributes:
        factory (Callable[[], Any] | None): The function used to create the default value for the attribute.
            Defaults to None.
        value (Any | None): The default value for the attribute. Defaults to None.
    """

    factory: Annotated[
        Callable[..., Any] | None,
        Field(default=None, description="The function used to create the default value for the attribute"),
    ]
    value: Annotated[Any | None, Field(default=None, description="The default value for the attribute")]


@dataclass
class FieldDefinition:
    """Represents a field with its metadata.

    Attributes:
        name (str): Name of the field.
        type (Any): Type annotation of the field.
        default (Default | None): Default value/factory of the field. Defaults to None.
        description (str | None): Description of the field. Defaults to None.
        optional (bool | None): Indicates if the field accepts None. Defaults to None.
        title (str | None): Title of the field. Defaults to None.
    """

    name: Annotated[str, Field(description="Name of the field")]
    type: Annotated[Any, Field(description="Type annotation of the field")]
    default: Annotated[Default | None, Field(default=None, description="Default value/factory of the field")]
    description: Annotated[str | None, Field(default=None, description="Description of the field")]
    optional: Annotated[bool | None, Field(default=None, description="Indicates if the field accepts None")]
    title: Annotated[str | None, Field(default=None, description="Title of the field")]

    @property
    def _metadata(self) -> dict[str, Any]:
        result: dict[str, Any] = {
            "description": self.description or f"Field {self.name}",
            "title": self.title or self.name,
        }
        if self.default:
            if self.default.factory:
                result["default_factory"] = self.default.factory
            else:
                result["default"] = self.default.value
        return result

    @property
    def type_annotated(self) -> Annotated[Any, Field]:
        """Returns the type annotation of the field.

        Returns:
            Annotated[Any, Field]: The type annotation of the field.
        """
        field_type = (self.type, self.type | None)[bool(self.optional)]
        return Annotated[field_type, Field(**self._metadata)]


class Base(BaseModel):
    """Represents a base class for models."""


class BaseWithId(Base):
    """Represents a base class for models with a resource identifier."""

    id: Annotated[UUID, Field(description="Resource identifier")]


T = TypeVar("T")


class Models:
    """Represents a collection of models.

    This class is responsible for creating and managing models dynamically based on the provided fields.
    It provides methods to create different types of models such as create, response, and update models.

    Attributes:
        name (str): The name of the collection of models.
        create_model (type[Base]): The dynamically created model for creating new instances.
        response_model (type[BaseWithId]): The dynamically created model for response payloads.
        update_model (type[Base]): The dynamically created model for updating existing instances.
    """

    def __init__(self, name: str, fields: list[FieldDefinition]) -> None:
        """Initialize the Models class.

        Args:
            name (str): The name of the collection of models.
            fields (list[FieldDefinition]): The list of fields definition for the models.
        """
        self.name = name
        self.create_model = self._factory(Base, f"Create{name.title()}", fields)
        self.response_model = self._factory(BaseWithId, f"Response{name.title()}", fields)
        self.update_model = self._factory(Base, f"Update{name.title()}", fields)

    @staticmethod
    def _factory(base: T, name: str, fields: list[FieldDefinition]) -> T:
        fields_constructor = {field.name: field.type_annotated for field in fields}
        return create_model(name, __base__=base, **fields_constructor)  # type: ignore



"""This module provides a high-level interface for interacting with different database clients."""
from importlib import import_module
from typing import Annotated, Any, Literal
from uuid import UUID

from fastapi import HTTPException, status
from pydantic import Field
from pydantic.dataclasses import dataclass

from ..dependencies import PaginateParameters

ResourceId = Annotated[UUID, Field(description="The unique identifier of the resource")]
ResourceObj = Annotated[dict[str, Any], Field(description="The resource object")]
Partial = Annotated[bool | None, Field(description="Perform a partial update")]


class Base:
    """Base client class for interacting with a database.

    Provides the basic interface for performing CRUD operations on a database.
    Derived classes should implement the specific methods for each operation.

    Attributes:
        exception (HTTPException): The exception module to be used for raising HTTP exceptions.
        status (status): The status module to be used for HTTP status codes.

    Args:
        path (str): The path to the database.
        table (str): The name of the table in the database.
    """

    exception = HTTPException
    status = status

    def __init__(self, path: str, table: str) -> None:
        """Initialize the Base class.

        Args:
            path (str): The path to the database.
            table (str): The name of the table in the database.
        """

    async def insert_one(self, resource_id: ResourceId, resource_obj: ResourceObj) -> ResourceObj:
        """Inserts a resource object in the database.

        Args:
            resource_id (ResourceId): The ID of the resource to be inserted.
            resource_obj (ResourceObj): The resource object to be inserted.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource already exists in the database.
        """
        raise NotImplementedError

    async def select_one(self, resource_id: ResourceId) -> ResourceObj:
        """Retrieves a resource from the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be retrieved.

        Returns:
            ResourceObj: The retrieved resource object.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        raise NotImplementedError

    async def update_one(
        self, resource_id: ResourceId, resource_obj: ResourceObj, partial: Partial = None
    ) -> ResourceObj:
        """Updates a resource in the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be updated.
            resource_obj (ResourceObj): The resource object to be updated.
            partial (Partial): Whether to perform a partial update or replace the entire resource object.
                Defaults to None.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        raise NotImplementedError

    async def delete_one(self, resource_id: ResourceId) -> None:
        """Deletes a resource from the database based on the given resource ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be deleted.

        Returns:
            None

        Raises:
            self.exception: If the resource is not found in the database.
        """
        raise NotImplementedError

    async def select_many(self, paginate_parameters: PaginateParameters) -> list[ResourceObj]:
        """Retrieves multiple resources from the database based on the given pagination parameters.

        Args:
            paginate_parameters (PaginateParameters): The pagination parameters.

        Returns:
            list[ResourceObj]: A list of objects representing the retrieved resources.
        """
        raise NotImplementedError


@dataclass
class DBSpecification:
    """Represents the specification for a database connection.

    Attributes:
        path (str): Path to the database.
            This field specifies the path to the database file or server.
        client (str): One of available database clients.
            This field specifies the database client to be used for the connection.
            The available options are "Default", "Postgres", "Redis", "SQLite", and "TinyDB".
            Defaults to "Default".
    """

    path: Annotated[str, Field(description="Path to the database")]
    client: Annotated[
        Literal["Default", "Postgres", "Redis", "SQLite", "TinyDB"],
        Field(default="Default", description="One of available database clients"),
    ]


class DB:
    """Represents a database client adapter.

    Attributes:
        client (type[Base]): The underlying client used for specific database operations.

    Methods:
        insert_one(self, *args: Any, **kwargs: Any) -> Any:
            Inserts a single record into the database.

        select_one(self, *args: Any, **kwargs: Any) -> Any:
            Retrieves a single record from the database.

        update_one(self, *args: Any, **kwargs: Any) -> Any:
            Updates a single record in the database.

        delete_one(self, *args: Any, **kwargs: Any) -> Any:
            Deletes a single record from the database.

        select_many(self, *args: Any, **kwargs: Any) -> Any:
            Retrieves multiple records from the database.
    """

    client: Base

    def __init__(self, db_spec: DBSpecification, table: str) -> None:
        """Initialize the DB instance.

        Args:
            db_spec (DBSpecification): The specification for the database connection.
            table (str): The name of the table to interact with.
        """
        client_module = import_module(f".{db_spec.client.lower()}", package=__package__)
        client_class: type[Base] = getattr(client_module, f"{db_spec.client}Client")
        self.client = client_class(db_spec.path, table)

    async def insert_one(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> ResourceObj:
        """Insert a single record into the database.

        Args:
            *args: Positional arguments to be passed to the client's insert_one method.
            **kwargs: Keyword arguments to be passed to the client's insert_one method.

        Returns:
            Coroutine: A coroutine representing the asynchronous insertion operation.
        """
        return await self.client.insert_one(*args, **kwargs)

    async def select_one(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> ResourceObj:
        """Select a single record from the database.

        Args:
            *args: Positional arguments to be passed to the client's select_one method.
            **kwargs: Keyword arguments to be passed to the client's select_one method.

        Returns:
            Coroutine: A coroutine representing the asynchronous selection operation.
        """
        return await self.client.select_one(*args, **kwargs)

    async def update_one(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> ResourceObj:
        """Update a single record in the database.

        Args:
            *args: Positional arguments to be passed to the client's update_one method.
            **kwargs: Keyword arguments to be passed to the client's update_one method.

        Returns:
            Coroutine: A coroutine representing the asynchronous update operation.
        """
        return await self.client.update_one(*args, **kwargs)

    async def delete_one(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Delete a single record from the database.

        Args:
            *args: Positional arguments to be passed to the client's delete_one method.
            **kwargs: Keyword arguments to be passed to the client's delete_one method.

        Returns:
            Coroutine: A coroutine representing the asynchronous deletion operation.
        """
        return await self.client.delete_one(*args, **kwargs)

    async def select_many(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> list[ResourceObj]:
        """Select multiple records from the database.

        Args:
            *args: Positional arguments to be passed to the client's select_many method.
            **kwargs: Keyword arguments to be passed to the client's select_many method.

        Returns:
            Coroutine: A coroutine representing the asynchronous selection operation.
        """
        return await self.client.select_many(*args, **kwargs)



"""This module provides the client class for interacting with a TinyDB database."""
from tinydb import Query, TinyDB

from . import Base, PaginateParameters, Partial, ResourceId, ResourceObj


class TinyDBclient(Base):
    """A class representing a TinyDB client for database operations.

    Args:
        path (str): The path to the TinyDB database file.
        table (str): The name of the table in the database.

    Attributes:
        db (TinyDB): The TinyDB instance.
        path (str): The path to the TinyDB database file.
        table (str): The name of the table in the database.
    """

    db: TinyDB
    path: str
    table: str

    def __init__(self, path: str, table: str) -> None:
        """Initialize the TinyDBclient class.

        Args:
            path (str): The path to the database.
            table (str): The name of the table in the database.

        """
        self.db = TinyDB(path)
        self.table = table

    def _get(self, resource_id: ResourceId) -> ResourceObj:
        """Retrieves a resource from the database based on the given resource ID.

        Args:
            resource_id (ResourceId): The ID of the resource to retrieve.

        Returns:
            ResourceObj: The retrieved resource.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        try:
            result = self.db.table(self.table).search(Query().resource_id == str(resource_id))  # type: ignore
            return result[0]
        except (KeyError, IndexError) as exception:
            raise self.exception(self.status.HTTP_404_NOT_FOUND, "not found") from exception

    async def insert_one(self, resource_id: ResourceId, resource_obj: ResourceObj) -> ResourceObj:
        """Inserts a resource object in the database.

        Args:
            resource_id (ResourceId): The ID of the resource to be inserted.
            resource_obj (ResourceObj): The resource object to be inserted.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource already exists in the database.
        """
        self.db.table(self.table).insert({"resource_id": str(resource_id), "resource_obj": resource_obj})  # type: ignore
        return {"id": resource_id, **resource_obj}

    async def select_one(self, resource_id: ResourceId) -> ResourceObj:
        """Retrieves a resource from the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be retrieved.

        Returns:
            ResourceObj: The retrieved resource object.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        result = self._get(resource_id)
        return {"id": resource_id, **result["resource_obj"]}

    async def update_one(
        self, resource_id: ResourceId, resource_obj: ResourceObj, partial: Partial = None
    ) -> ResourceObj:
        """Updates a resource in the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be updated.
            resource_obj (ResourceObj): The resource object to be updated.
            partial (Partial): Whether to perform a partial update or replace the entire resource object.
                Defaults to None.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        _resource_obj = self._get(resource_id)
        if partial:
            _resource_obj.update(resource_obj)
        else:
            _resource_obj = resource_obj
        self.db.table(self.table).update({"resource_obj": _resource_obj}, Query().resource_id == str(resource_id))  # type: ignore
        return {"id": resource_id, **_resource_obj}

    async def delete_one(self, resource_id: ResourceId) -> None:
        """Deletes a resource from the database based on the given resource ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be deleted.

        Returns:
            None

        Raises:
            self.exception: If the resource is not found in the database.
        """
        self._get(resource_id)
        self.db.table(self.table).remove(Query().resource_id == str(resource_id))  # type: ignore

    async def select_many(self, paginate_parameters: PaginateParameters) -> list[ResourceObj]:
        """Retrieves multiple resources from the database based on the given pagination parameters.

        Args:
            paginate_parameters (PaginateParameters): The pagination parameters.

        Returns:
            list[ResourceObj]: A list of objects representing the retrieved resources.
        """
        return [{"id": result["resource_id"], **result["resource_obj"]} for result in self.db.table(self.table).all()][
            paginate_parameters.skip : paginate_parameters.limit
        ]  # type: ignore



from unittest import IsolatedAsyncioTestCase
from uuid import uuid4

from sthali_db.models import Any, Base, BaseWithId, Default, FieldDefinition, Models


class TestDefault(IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        result = Default()  # type: ignore

        self.assertEqual(result.factory, None)
        self.assertEqual(result.value, None)

    async def test_return_custom(self) -> None:
        def func() -> None:
            return

        result = Default(factory=func, value=0)

        self.assertEqual(result.factory, func)
        self.assertEqual(result.value, 0)


class TestFieldDefinition(IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        result = FieldDefinition(name="name", type=Any)  # type: ignore

        self.assertEqual(result.name, "name")
        self.assertEqual(result.type, Any)
        self.assertEqual(result.default, None)
        self.assertEqual(result.description, None)
        self.assertEqual(result.optional, None)
        self.assertEqual(result.title, None)

    async def test_return_custom(self) -> None:
        def func() -> None:
            return

        result = FieldDefinition(
            name="name",
            type=str,
            default={"factory": func, "value": 0},  # type: ignore
            description="description",
            optional=True,
            title="title",
        )

        self.assertEqual(result.name, "name")
        self.assertEqual(result.type, str)
        self.assertEqual(result.default.factory, func)  # type: ignore
        self.assertEqual(result.default.value, 0)  # type: ignore
        self.assertEqual(result.description, "description")
        self.assertEqual(result.optional, True)
        self.assertEqual(result.title, "title")

    async def test_type_annotated(self) -> None:
        field_definition = FieldDefinition(name="name", type=str)  # type: ignore

        result = field_definition.type_annotated

        self.assertEqual(result("str"), "str")
        self.assertEqual(result.__args__[0], str)
        self.assertEqual(result.__metadata__[0].title, "name")

    async def test_type_annotated_with_optional(self) -> None:
        field_definition = FieldDefinition(name="name", type=str, optional=True)  # type: ignore

        result = field_definition.type_annotated

        self.assertEqual(result.__args__[0], str | None)

    async def test_type_annotated_with_default_factory(self) -> None:
        def func() -> None:
            return

        field_definition = FieldDefinition(name="name", type=str, default={"factory": func, "value": 0})  # type: ignore

        result = field_definition.type_annotated

        self.assertEqual(result.__metadata__[0].default_factory(), None)

    async def test_type_annotated_with_default_value(self) -> None:
        field_definition = FieldDefinition(name="name", type=str, default={"value": 0})  # type: ignore

        result = field_definition.type_annotated

        self.assertEqual(result.__metadata__[0].default, 0)


class TestBase(IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        result = Base()

        self.assertEqual(result.model_dump(), {})


class TestBaseWithId(IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        _id = uuid4()
        result = BaseWithId(id=_id)

        self.assertEqual(result.model_dump(), {"id": _id})


class TestModels(IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        _id = uuid4()

        result = Models(name="name", fields=[])

        self.assertEqual(result.name, "name")
        self.assertEqual(result.create_model().model_dump(), {})
        self.assertEqual(result.response_model(**{"id": _id}).model_dump(), {"id": _id})
        self.assertEqual(result.update_model().model_dump(), {})

    async def test_return_custom(self) -> None:
        def func() -> None:
            return

        _id = uuid4()

        result = Models(
            name="name",
            fields=[
                FieldDefinition(name="field1", type=str),  # type: ignore
                FieldDefinition(name="field2", type=str, optional=True),  # type: ignore
                FieldDefinition(name="field3", type=str, default={"factory": func}),  # type: ignore
                FieldDefinition(name="field4", type=str, default={"value": "field4"}),  # type: ignore
                FieldDefinition(name="field5", type=str, default={"factory": func, "value": 1}),  # type: ignore
            ],
        )

        self.assertEqual(result.name, "name")
        self.assertEqual(
            result.create_model(**{"field1": "field1", "field2": None}).model_dump(),
            {"field1": "field1", "field2": None, "field3": None, "field4": "field4", "field5": None},
        )
        self.assertEqual(
            result.response_model(**{"id": _id, "field1": "field1", "field2": None}).model_dump(), # type: ignore
            {"id": _id, "field1": "field1", "field2": None, "field3": None, "field4": "field4", "field5": None},
        )
        self.assertEqual(
            result.update_model(**{"field1": "field1", "field2": None}).model_dump(),
            {"field1": "field1", "field2": None, "field3": None, "field4": "field4", "field5": None},
        )
